// this file contains all the initializations and code required to process a single HET A packet (ApID = 590)
// in particular, it contains a subroutine 'int ProcessAPacket(unsigned char *ubuf)'
// which processes a single, binary A Packet in ubuf[].  ubuf includes the ground receipt header (from which
// the SCindex is determined (0 for Ahead and 1 for Behind)), the CCSDS header (from which the time is determined),
// and the HET packet data. A_PacketOut contains the results of processing as a binary structure (124 bytes)
// A_PacketOut is extern to main()
//
// 25 Oct 2009 - Revised SWCtrX[109] to NSWCtrX[109] (log-compressed counts in each SW counter) and SWCtr[109] to NSWCtr[109] 
//               (uncompressed counts) ... revised names are clearer as to what these variables actually refer to

#include <stdio.h>
#include <stdlib.h>
#include "PacketAeH.h"

#include "PacketA_init.h"
// contains quantities needed to unpack HET A Packet fluxes:
//      #defines
//      int swctr[27][NXXBINSMAX];			        // the software counter corresponding to [prtcltype][KE bin index for this particle type]
//      float deltaKE[SCindex][27][NXXBINSMAX];		// deltaKE[prtcltype][KE bin index]
//      float GF[27][2][NXXBINSMAX];				// [prtcltype][high_rate][KE bin index]
//
//      Note that the deltaKE[][][] array includes the SCindex whereas the swctr[][] and GF[][][] arrays do not

#include <math.h>

#define NPACKETBYTES   261
// the Ground Receipt header is an additional 26 bits    
// the CCSDS header is an additional 11 bits
// the data is 261 bits

#define SECONDS_OFFSET  -378691200L  
     // 1958 Jan 1 00:00:00 relative to 
	 // 1970 Jan 1 00:00:00

unsigned char SC_ID;          // spacecraft ID  0xEA = Ahead, 0xEB = Behind
int SCindex;				  // 0 for Ahead, 1 for Behind

struct uncomp_hdr_CCSDS     // CCSDS hdr 
{
	unsigned short version; 
	unsigned short type;
	unsigned short hdr_flg;
	short ApId;
	unsigned short grp_flg;
	short seq_cnt;
	unsigned short pck_len;
	int secs;
	unsigned short msecs;     // millisecs
}; 
 
struct uncomp_hdr_CCSDS *puncomp_CCSDShdr,uncomp_CCSDShdr;

struct PacketA			// 590 (dec)  24e (hex) ... this struct reflects the A packet image formed in memory (only one TBD0 byte) by unpack_APacket()
{  // A data packet without ground receipt header or CCSDS header
	unsigned char    HET_mode;
	unsigned char    TBD0;		   // in the input data packet there are actually 2-bytes for TBD0, one of which will be discarded 
								   // using only 1 here avoids a pad byte in the structure
								   // Note: the data is read into buffer as 261 bytes
	unsigned short   MajorFrame;   // not log-compressed
	unsigned short   LivetimeX;    // X => log-compressed
	unsigned short   NTriggersX;
	unsigned short   NCoincX;
	unsigned short	 NTotalEventsX;
	unsigned short   NSinglesQdX;
	unsigned short   NStopQdX;
	unsigned short   NPenQdX;
	unsigned short   NStopHX;
	unsigned short   NStopHeX;
	unsigned short   NStopZgt2X;
	unsigned short   NPenHX;
	unsigned short   NPenHeX;
	unsigned short   NPenZgt2X;
	unsigned short   NInvalidX[4];
	unsigned short   NStimX;
	unsigned short   NSWCtrX[109];   // log-compressed counts
	unsigned char    TBD1;
	unsigned char    checksum;   
};

struct PacketA  A_Packet,*pA_Packet;  // buffer of packet data from Peter S. is converted to structure A_Packet by unpack_APacket
									  // A_Packet is defined/declared here and main() doesn't know about it
									  
struct Flux
{
	float flux;
	float sigma;
};

struct PacketAOut
{
	long	tsec70;
	int	    high_rate;
	int     SCindex;         // 0 for Ahead, 1 for Behind
    struct  Flux flux[16];   // SWCtrs 6 - 11,13-18, and 81-84
};

struct PacketAOut  A_PacketOut, *pA_PacketOut; // 18 Jan 2010


int fluxindex;

unsigned char    HET_mode;
int              high_rate;			// used to index into an array
unsigned short   MajorFrame;

float	Livetime;			// seconds
float	NTriggers;		    // counts
float	NCoinc;			    // counts
float	NTotalEvents;		
float	NSinglesQd;
float	NStopQd;
float	NPenQd;
float	NStopH;
float	NStopHe;
float	NStopZgt2;
float	NPenH;
float	NPenHe;
float	NPenZgt2;
float	NInvalid[4];
float	NStim;
float	NSWCtr[109];          // uncompressed counts!

long t_packet58;

float float_rate(unsigned short packed); 
// Unpack to float ... Macro handles floats (4 bytes) but not doubles (8 bytes)

void uncompress_CCSDShdr(unsigned char* ubuf);
// Given the first 11 bytes (CCSDS header), this routine will uncompress the values into a structure
// The CCSDS header has the bytes in big-endian order
// The code should work equally well on big-endian and little-endian machines

void unpack_APacket(unsigned char *ubuf);  // unpacks buffer of data from Peter S. to a structure defined/declared in PacketAeH.c

int ProcessAPacket(unsigned char *ubuf);
//		ubuf is a pointer to the buffer which contains the binary A packet to be processed
//      (it is assumed that the packet has a Ground Receipt header (26 bits), a CCSDS header (11 bits) and the data (261 bytes)
//		pA_PacketOut is a pointer to the location where the subroutine puts the processed data
//      this location and pointer are defined in PacketAeH.c; the pointer is accessed by main() via an extern statement 
//
// Returns 1 if successful, else 0

float float_rate(unsigned short packed) /* Unpack to float ... Macro handles floats (4 bytes) but not doubles (8 bytes) */
{
        int power;
        float out;

        power= packed>>11;
        if (power>1)
        {
                out=(float)((packed&0x07ff)|0x0800);
                out=out*(float)pow(2.,power-1);
        }
        else
                out=packed;
        return out;
}




// Given the first 11 bytes (CCSDS header), this routine will uncompress the values into a structure
// The CCSDS header has the bytes in big-endian order
// The code should work equally well on big-endian and little-endian machines
void uncompress_CCSDShdr(unsigned char* ubuf)
{
	 puncomp_CCSDShdr->version = ubuf[0] >> 5;
	 puncomp_CCSDShdr->type = (ubuf[0] & 16) >> 4;
	 puncomp_CCSDShdr->hdr_flg = (ubuf[0] & 8) >> 3;

		// shft into the MSB and then add the LSB and mask off bits
	 puncomp_CCSDShdr->ApId = ((ubuf[0] << 8) + ubuf[1]) & 0x07ff;
	 puncomp_CCSDShdr->grp_flg = ubuf[2] >> 6; 
	 puncomp_CCSDShdr->seq_cnt = ((ubuf[2] << 8) + ubuf[3]) & 0x3fff;

	 puncomp_CCSDShdr->pck_len = ((ubuf[4] << 8) + ubuf[5]);
 	 puncomp_CCSDShdr->secs =	(ubuf[6] << 24) + (ubuf[7] << 16) + (ubuf[8] << 8) + ubuf[9];
	 puncomp_CCSDShdr->msecs = ubuf[10]*4;
	 
	 if (puncomp_CCSDShdr->msecs > 500) 
		puncomp_CCSDShdr->secs++;
}



void unpack_APacket(unsigned char *ubuf)
{ 
	 int i;

//	 printf("\nubuf[2]=0x%x, ubuf[3]=0x%x",ubuf[2],ubuf[3]);

	 pA_Packet->HET_mode = ubuf[0];					// November 27, 2006
	 HET_mode = pA_Packet->HET_mode;
	 high_rate = (int)((HET_mode & 16)>>4);     // November 27, 2006
	 pA_Packet->MajorFrame=(ubuf[4] << 8)+ubuf[3];   // not log-compressed
	 MajorFrame = pA_Packet->MajorFrame;

	 pA_Packet->LivetimeX=(ubuf[6] << 8) + ubuf[5];
	 Livetime=float_rate(pA_Packet->LivetimeX)*(float)5e-6;		// seconds

	 pA_Packet->NTriggersX=(ubuf[8] << 8) + ubuf[7];	
	 NTriggers=float_rate(pA_Packet->NTriggersX);		// counts

	 pA_Packet->NCoincX=(ubuf[10] << 8) + ubuf[9];
	 NCoinc=float_rate(pA_Packet->NCoincX);			// counts

	 pA_Packet->NTotalEventsX=(ubuf[12] << 8) + ubuf[11];	
	 NTotalEvents=float_rate(pA_Packet->NTotalEventsX);	// counts

	 pA_Packet->NSinglesQdX=(ubuf[14] << 8) + ubuf[13];
	 NSinglesQd=float_rate(pA_Packet->NSinglesQdX);	// counts

	 pA_Packet->NStopQdX=(ubuf[16] << 8) + ubuf[15];
	 NStopQd=float_rate(pA_Packet->NStopQdX);

	 pA_Packet->NPenQdX=(ubuf[18] << 8) + ubuf[17];
	 NPenQd=float_rate(pA_Packet->NPenQdX);

	 pA_Packet->NStopHX=(ubuf[20] << 8) + ubuf[19];
	 NStopH=float_rate(pA_Packet->NStopHX);

	 pA_Packet->NStopHeX=(ubuf[22] << 8) + ubuf[21];
	 NStopHe=float_rate(pA_Packet->NStopHeX);

	 pA_Packet->NStopZgt2X=(ubuf[24] << 8) + ubuf[23];
	 NStopZgt2=float_rate(pA_Packet->NStopZgt2X);

	 pA_Packet->NPenHX=(ubuf[26] << 8) + ubuf[25];
	 NPenH=float_rate(pA_Packet->NPenHX);

	 pA_Packet->NPenHeX=(ubuf[28] << 8) + ubuf[27];
	 NPenHe=float_rate(pA_Packet->NPenHeX);

	 pA_Packet->NPenZgt2X=(ubuf[30] << 8) + ubuf[29];
	 NPenZgt2=float_rate(pA_Packet->NPenZgt2X);

	 for(i=0;i<4;i++)
	 {
		pA_Packet->NInvalidX[i]=(ubuf[32+2*i] << 8) + ubuf[31+2*i];
		NInvalid[i]=float_rate(pA_Packet->NInvalidX[i]);
	 }
	 
	 pA_Packet->NStimX=(ubuf[40] << 8) + ubuf[39];
	 NStim=float_rate(pA_Packet->NStimX);

	 for(i=0;i<109;i++)
	 {
		 pA_Packet->NSWCtrX[i]=(ubuf[42+2*i] << 8) + ubuf[41+2*i];
		 NSWCtr[i]=float_rate(pA_Packet->NSWCtrX[i]);
	 }

//	 ubuf[259] = TBD1 (single byte)

	 pA_Packet->checksum=ubuf[260];
}



int ProcessAPacket(unsigned char *ubuf)
//		ubuf is a pointer to the buffer which contains the binary A packet to be processed
//      (it is assumed that the packet has a Ground Receipt header (26 bits), a CCSDS header (11 bits) and the data (261 bytes)
//		pA_PacketOut is a pointer to the location where the subroutine will put the processed data
//      pA_PacketOut is available to main() via an extern statement
//
// Returns 1 if successful, else 0
{	
		int prtcltype, bin;     // 10 Jan 2010
		SC_ID = ubuf[6];		// 0xEA for Spacecraft A, 0xEB for Spacecraft B

//printf("Here 1\n");
		if(SC_ID == 0xEA)
			SCindex = 0;
		else if(SC_ID == 0xEB)
			SCindex = 1;
		else
		{// this should never happen
			printf("\nError in ProcessAPacket(): SC_ID=0x%x (expect 0xEA or 0xEB) ",SC_ID); 
			printf("Hit Enter to exit: ");
			getchar();
			exit(1);
		}
//printf("Here 2\n");
		puncomp_CCSDShdr=&uncomp_CCSDShdr;
		uncompress_CCSDShdr(&ubuf[26]);

		if((puncomp_CCSDShdr->ApId) != (short)590)   
			return 0;   // not a HET A Packet

		t_packet58 = puncomp_CCSDShdr->secs;

		pA_Packet=&A_Packet;
		pA_PacketOut = &A_PacketOut;
//printf("Here 3\n");

		pA_PacketOut->tsec70 = t_packet58 + SECONDS_OFFSET;
// debug
// printf("\n\ndebug ProcessAPacket in PacketAeH.h:");
// printf("\nSECONDS_OFFSET = %ld, t_packet58 = %ld; Enter to cont\n\n",SECONDS_OFFSET,t_packet58);
// getchar();

		unpack_APacket(&ubuf[37]);  // sets high_rate & Livetime

		pA_PacketOut->high_rate = high_rate;

		pA_PacketOut->SCindex = SCindex;

		fluxindex = 0;
//printf("Here 4\n");

// electrons  PrtclType = 1

// printf("Electrons %d\n",NELECTRONBINS);
		prtcltype=1;
		for(bin=0;bin<NELECTRONBINS;bin++)
		{
			pA_PacketOut->flux[fluxindex].flux = NSWCtr[swctr[prtcltype][bin]]/(GF[prtcltype][high_rate][bin]*Livetime*deltaKE[SCindex][prtcltype][bin]);
			pA_PacketOut->flux[fluxindex].sigma = (float)sqrt(NSWCtr[swctr[prtcltype][bin]])/(GF[prtcltype][high_rate][bin]*Livetime*deltaKE[SCindex][prtcltype][bin]);
			fluxindex++;
		}

// protons    PrtclType = 2

// printf("Protons %d\n",NPROTONBINS);
		prtcltype=2;
		for(bin=0;bin<NPROTONBINS;bin++)
		{
// printf("bin= %d fluxindex= %d, high_rate= %o\n",bin,fluxindex,high_rate);
			if(swctr[prtcltype][bin] == 12)
				continue;  // omits protons from 19.0 - 28.0 MeV
			else
			{
// printf("bin= %d fluxindex= %d, high_rate= %o\n",bin,fluxindex,high_rate);

				pA_PacketOut->flux[fluxindex].flux = NSWCtr[swctr[prtcltype][bin]]/(GF[prtcltype][high_rate][bin]*Livetime*deltaKE[SCindex][prtcltype][bin]);

//printf("Before bin= %d fluxindex= %d, high_rate= %o\n",bin,fluxindex,high_rate);
//printf("Flux = %f sigma = %f\n", NSWCtr[swctr[prtcltype][bin]]/(GF[prtcltype][high_rate][bin]*Livetime*deltaKE[SCindex][prtcltype][bin]), (float)sqrt(NSWCtr[swctr[prtcltype][bin]])/(GF[prtcltype][high_rate][bin]*Livetime*deltaKE[SCindex][prtcltype][bin]));
//printf("prtcltype = %d bin = %d swctr[prtclytype][bin]] =%d\n", prtcltype,bin,swctr[prtcltype][bin]);
//printf("GF =  %f SCindex = %d deltaKE = %f\n",GF[prtcltype][high_rate][bin],SCindex,deltaKE[SCindex][prtcltype][bin]);
//printf("pA_PacketOut->flux[fluxindex].sigma before = %f\n",pA_PacketOut->flux[fluxindex].sigma);

				pA_PacketOut->flux[fluxindex].sigma = (float)sqrt(NSWCtr[swctr[prtcltype][bin]])/(GF[prtcltype][high_rate][bin]*Livetime*deltaKE[SCindex][prtcltype][bin]);

//printf("After bin= %d fluxindex= %d, high_rate= %o\n",bin,fluxindex,high_rate);
//printf("pA_PacketOut->flux[fluxindex].sigma after = %f\n",pA_PacketOut->flux[fluxindex].sigma);
				fluxindex++;

			}
		}

//printf("Here 5\n");
		return 1;
}
